@@ -1,12 +1,18 @@
 """Collection of language-specific analyzers and analyzer registry for Annif"""
 
+from __future__ import annotations
+
 import re
+from typing import TYPE_CHECKING
 
 import annif
 from annif.util import parse_args
 
 from . import simple, simplemma, snowball
 
+if TYPE_CHECKING:
+    from annif.analyzer.analyzer import Analyzer
+
 _analyzers = {}
 
 
@@ -14,7 +20,7 @@ def register_analyzer(analyzer):
     _analyzers[analyzer.name] = analyzer
 
 
-def get_analyzer(analyzerspec):
+def get_analyzer(analyzerspec: str) -> Analyzer:
     match = re.match(r"(\w+)(\((.*)\))?", analyzerspec)
     if match is None:
         raise ValueError("Invalid analyzer specification {}".format(analyzerspec))
