@@ -1,5 +1,7 @@
 """Snowball analyzer for Annif, based on nltk Snowball stemmer."""
 
+from __future__ import annotations
+
 import functools
 
 from . import analyzer
@@ -8,7 +10,7 @@ from . import analyzer
 class SnowballAnalyzer(analyzer.Analyzer):
     name = "snowball"
 
-    def __init__(self, param, **kwargs):
+    def __init__(self, param: str, **kwargs) -> None:
         self.param = param
         import nltk.stem.snowball
 
@@ -16,5 +18,5 @@ class SnowballAnalyzer(analyzer.Analyzer):
         super().__init__(**kwargs)
 
     @functools.lru_cache(maxsize=500000)
-    def _normalize_word(self, word):
+    def _normalize_word(self, word: str) -> str:
         return self.stemmer.stem(word.lower())
