@@ -1,5 +1,7 @@
 """Voikko analyzer for Annif, based on libvoikko library."""
 
+from __future__ import annotations
+
 import functools
 
 import voikko.libvoikko
@@ -10,12 +12,12 @@ from . import analyzer
 class VoikkoAnalyzer(analyzer.Analyzer):
     name = "voikko"
 
-    def __init__(self, param, **kwargs):
+    def __init__(self, param: str, **kwargs) -> None:
         self.param = param
         self.voikko = None
         super().__init__(**kwargs)
 
-    def __getstate__(self):
+    def __getstate__(self) -> dict[str, str | None]:
         """Return the state of the object for pickling purposes. The Voikko
         instance is set to None because as a ctypes object it cannot be
         pickled."""
@@ -23,7 +25,7 @@ class VoikkoAnalyzer(analyzer.Analyzer):
         return {"param": self.param, "voikko": None}
 
     @functools.lru_cache(maxsize=500000)
-    def _normalize_word(self, word):
+    def _normalize_word(self, word: str) -> str:
         if self.voikko is None:
             self.voikko = voikko.libvoikko.Voikko(self.param)
         result = self.voikko.analyze(word)
