@@ -1,7 +1,10 @@
 """Annif backend using the fastText classifier"""
 
+from __future__ import annotations
+
 import collections
 import os.path
+from typing import TYPE_CHECKING, Any
 
 import fasttext
 
@@ -11,6 +14,12 @@ from annif.suggestion import SubjectSuggestion
 
 from . import backend, mixins
 
+if TYPE_CHECKING:
+    from fasttext.FastText import _FastText
+    from numpy import ndarray
+
+    from annif.corpus.document import DocumentCorpus
+
 
 class FastTextBackend(mixins.ChunkingBackend, backend.AnnifBackend):
     """fastText backend for Annif"""
@@ -48,14 +57,14 @@ class FastTextBackend(mixins.ChunkingBackend, backend.AnnifBackend):
     # defaults for uninitialized instances
     _model = None
 
-    def default_params(self):
+    def default_params(self) -> dict[str, Any]:
         params = backend.AnnifBackend.DEFAULT_PARAMETERS.copy()
         params.update(mixins.ChunkingBackend.DEFAULT_PARAMETERS)
         params.update(self.DEFAULT_PARAMETERS)
         return params
 
     @staticmethod
-    def _load_model(path):
+    def _load_model(path: str) -> _FastText:
         # monkey patch fasttext.FastText.eprint to avoid spurious warning
         # see https://github.com/facebookresearch/fastText/issues/1067
         orig_eprint = fasttext.FastText.eprint
@@ -65,7 +74,7 @@ class FastTextBackend(mixins.ChunkingBackend, backend.AnnifBackend):
         fasttext.FastText.eprint = orig_eprint
         return model
 
-    def initialize(self, parallel=False):
+    def initialize(self, parallel: bool = False) -> None:
         if self._model is None:
             path = os.path.join(self.datadir, self.MODEL_FILE)
             self.debug("loading fastText model from {}".format(path))
@@ -79,14 +88,14 @@ class FastTextBackend(mixins.ChunkingBackend, backend.AnnifBackend):
                 )
 
     @staticmethod
-    def _id_to_label(subject_id):
+    def _id_to_label(subject_id: int) -> str:
         return "__label__{:d}".format(subject_id)
 
-    def _label_to_subject_id(self, label):
+    def _label_to_subject_id(self, label: str) -> int:
         labelnum = label.replace("__label__", "")
         return int(labelnum)
 
-    def _write_train_file(self, corpus, filename):
+    def _write_train_file(self, corpus: DocumentCorpus, filename: str) -> None:
         with open(filename, "w", encoding="utf-8") as trainfile:
             for doc in corpus.documents:
                 text = self._normalize_text(doc.text)
@@ -98,17 +107,20 @@ class FastTextBackend(mixins.ChunkingBackend, backend.AnnifBackend):
                 else:
                     self.warning(f'no labels for document "{doc.text}"')
 
-    def _normalize_text(self, text):
+    def _normalize_text(self, text: str) -> str:
         return " ".join(self.project.analyzer.tokenize_words(text))
 
-    def _create_train_file(self, corpus):
+    def _create_train_file(
+        self,
+        corpus: DocumentCorpus,
+    ) -> None:
         self.info("creating fastText training file")
 
         annif.util.atomic_save(
             corpus, self.datadir, self.TRAIN_FILE, method=self._write_train_file
         )
 
-    def _create_model(self, params, jobs):
+    def _create_model(self, params: dict[str, Any], jobs: int) -> None:
         self.info("creating fastText model")
         trainpath = os.path.join(self.datadir, self.TRAIN_FILE)
         modelpath = os.path.join(self.datadir, self.MODEL_FILE)
@@ -123,7 +135,12 @@ class FastTextBackend(mixins.ChunkingBackend, backend.AnnifBackend):
         self._model = fasttext.train_supervised(trainpath, **params)
         self._model.save_model(modelpath)
 
-    def _train(self, corpus, params, jobs=0):
+    def _train(
+        self,
+        corpus: DocumentCorpus,
+        params: dict[str, Any],
+        jobs: int = 0,
+    ) -> None:
         if corpus != "cached":
             if corpus.is_empty():
                 raise NotSupportedException(
@@ -134,7 +151,9 @@ class FastTextBackend(mixins.ChunkingBackend, backend.AnnifBackend):
             self.info("Reusing cached training data from previous run.")
         self._create_model(params, jobs)
 
-    def _predict_chunks(self, chunktexts, limit):
+    def _predict_chunks(
+        self, chunktexts: list[str], limit: int
+    ) -> tuple[list[list[str]], list[ndarray]]:
         return self._model.predict(
             list(
                 filter(
@@ -144,7 +163,9 @@ class FastTextBackend(mixins.ChunkingBackend, backend.AnnifBackend):
             limit,
         )
 
-    def _suggest_chunks(self, chunktexts, params):
+    def _suggest_chunks(
+        self, chunktexts: list[str], params: dict[str, Any]
+    ) -> list[SubjectSuggestion]:
         limit = int(params["limit"])
         chunklabels, chunkscores = self._predict_chunks(chunktexts, limit)
         label_scores = collections.defaultdict(float)
