@@ -1,35 +1,52 @@
 """Neural network based ensemble backend that combines results from multiple
 projects."""
 
+from __future__ import annotations
 
+import importlib
+import json
 import os.path
 import shutil
+import zipfile
 from io import BytesIO
+from typing import TYPE_CHECKING, Any
 
 import joblib
+import keras
 import lmdb
 import numpy as np
-import tensorflow.keras.backend as K
+from keras.layers import Add, Dense, Dropout, Flatten, Input, Layer
+from keras.models import Model
+from keras.saving import load_model
+from keras.utils import Sequence
 from scipy.sparse import csc_matrix, csr_matrix
-from tensorflow.keras.layers import Add, Dense, Dropout, Flatten, Input, Layer
-from tensorflow.keras.models import Model, load_model
-from tensorflow.keras.utils import Sequence
 
 import annif.corpus
 import annif.parallel
 import annif.util
-from annif.exception import NotInitializedException, NotSupportedException
+from annif.exception import (
+    NotInitializedException,
+    NotSupportedException,
+    OperationFailedException,
+)
 from annif.suggestion import SuggestionBatch, vector_to_suggestions
 
 from . import backend, ensemble
 
+if TYPE_CHECKING:
+    from tensorflow.python.framework.ops import EagerTensor
 
-def idx_to_key(idx):
+    from annif.corpus.document import DocumentCorpus
+
+logger = annif.logger
+
+
+def idx_to_key(idx: int) -> bytes:
     """convert an integer index to a binary key for use in LMDB"""
     return b"%08d" % idx
 
 
-def key_to_idx(key):
+def key_to_idx(key: memoryview | bytes) -> int:
     """convert a binary LMDB key to an integer index"""
     return int(key)
 
@@ -47,7 +64,7 @@ class LMDBSequence(Sequence):
             self._counter = 0
         self._batch_size = batch_size
 
-    def add_sample(self, inputs, targets):
+    def add_sample(self, inputs: np.ndarray, targets: np.ndarray) -> None:
         # use zero-padded 8-digit key
         key = idx_to_key(self._counter)
         self._counter += 1
@@ -58,7 +75,7 @@ class LMDBSequence(Sequence):
         buf.seek(0)
         self._txn.put(key, buf.read())
 
-    def __getitem__(self, idx):
+    def __getitem__(self, idx: int) -> tuple[np.ndarray, np.ndarray]:
         """get a particular batch of samples"""
         cursor = self._txn.cursor()
         first_key = idx * self._batch_size
@@ -73,7 +90,7 @@ class LMDBSequence(Sequence):
             target_arrays.append(target_csr.toarray().flatten())
         return np.array(input_arrays), np.array(target_arrays)
 
-    def __len__(self):
+    def __len__(self) -> int:
         """return the number of available batches"""
         return int(np.ceil(self._counter / self._batch_size))
 
@@ -81,8 +98,8 @@ class LMDBSequence(Sequence):
 class MeanLayer(Layer):
     """Custom Keras layer that calculates mean values along the 2nd axis."""
 
-    def call(self, inputs):
-        return K.mean(inputs, axis=2)
+    def call(self, inputs: EagerTensor) -> EagerTensor:
+        return keras.ops.mean(inputs, axis=2)
 
 
 class NNEnsembleBackend(backend.AnnifLearningBackend, ensemble.BaseEnsembleBackend):
@@ -91,7 +108,7 @@ class NNEnsembleBackend(backend.AnnifLearningBackend, ensemble.BaseEnsembleBacke
 
     name = "nn_ensemble"
 
-    MODEL_FILE = "nn-model.h5"
+    MODEL_FILE = "nn-model.keras"
     LMDB_FILE = "nn-train.mdb"
 
     DEFAULT_PARAMETERS = {
@@ -106,12 +123,7 @@ class NNEnsembleBackend(backend.AnnifLearningBackend, ensemble.BaseEnsembleBacke
     # defaults for uninitialized instances
     _model = None
 
-    def default_params(self):
-        params = backend.AnnifBackend.DEFAULT_PARAMETERS.copy()
-        params.update(self.DEFAULT_PARAMETERS)
-        return params
-
-    def initialize(self, parallel=False):
+    def initialize(self, parallel: bool = False) -> None:
         super().initialize(parallel)
         if self._model is not None:
             return  # already initialized
@@ -126,11 +138,27 @@ class NNEnsembleBackend(backend.AnnifLearningBackend, ensemble.BaseEnsembleBacke
                 backend_id=self.backend_id,
             )
         self.debug("loading Keras model from {}".format(model_filename))
-        self._model = load_model(
-            model_filename, custom_objects={"MeanLayer": MeanLayer}
-        )
+        try:
+            self._model = load_model(
+                model_filename, custom_objects={"MeanLayer": MeanLayer}
+            )
+        except Exception as err:
+            metadata = self.get_model_metadata(model_filename)
+            keras_version = importlib.metadata.version("keras")
+            message = (
+                f"loading Keras model from {model_filename}; "
+                f"model metadata: {metadata}; "
+                f"you have Keras version {keras_version}. "
+                f'Original error message: "{err}"'
+            )
+            raise OperationFailedException(message, backend_id=self.backend_id)
 
-    def _merge_source_batches(self, batch_by_source, sources, params):
+    def _merge_source_batches(
+        self,
+        batch_by_source: dict[str, SuggestionBatch],
+        sources: list[tuple[str, float]],
+        params: dict[str, Any],
+    ) -> SuggestionBatch:
         src_weight = dict(sources)
         score_vectors = np.array(
             [
@@ -153,7 +181,7 @@ class NNEnsembleBackend(backend.AnnifLearningBackend, ensemble.BaseEnsembleBacke
             self.project.subjects,
         )
 
-    def _create_model(self, sources):
+    def _create_model(self, sources: list[tuple[str, float]]) -> None:
         self.info("creating NN ensemble model")
 
         inputs = Input(shape=(len(self.project.subjects), len(sources)))
@@ -185,7 +213,12 @@ class NNEnsembleBackend(backend.AnnifLearningBackend, ensemble.BaseEnsembleBacke
         self._model.summary(print_fn=summary.append)
         self.debug("Created model: \n" + "\n".join(summary))
 
-    def _train(self, corpus, params, jobs=0):
+    def _train(
+        self,
+        corpus: DocumentCorpus,
+        params: dict[str, Any],
+        jobs: int = 0,
+    ) -> None:
         sources = annif.util.parse_sources(self.params["sources"])
         self._create_model(sources)
         self._fit_model(
@@ -195,7 +228,12 @@ class NNEnsembleBackend(backend.AnnifLearningBackend, ensemble.BaseEnsembleBacke
             n_jobs=jobs,
         )
 
-    def _corpus_to_vectors(self, corpus, seq, n_jobs):
+    def _corpus_to_vectors(
+        self,
+        corpus: DocumentCorpus,
+        seq: LMDBSequence,
+        n_jobs: int,
+    ) -> None:
         # pass corpus through all source projects
         sources = dict(annif.util.parse_sources(self.params["sources"]))
 
@@ -236,7 +274,13 @@ class NNEnsembleBackend(backend.AnnifLearningBackend, ensemble.BaseEnsembleBacke
             shutil.rmtree(lmdb_path)
         return lmdb.open(lmdb_path, map_size=lmdb_map_size, writemap=True)
 
-    def _fit_model(self, corpus, epochs, lmdb_map_size, n_jobs=1):
+    def _fit_model(
+        self,
+        corpus: DocumentCorpus,
+        epochs: int,
+        lmdb_map_size: int,
+        n_jobs: int = 1,
+    ) -> None:
         env = self._open_lmdb(corpus == "cached", lmdb_map_size)
         if corpus != "cached":
             if corpus.is_empty():
@@ -256,8 +300,25 @@ class NNEnsembleBackend(backend.AnnifLearningBackend, ensemble.BaseEnsembleBacke
 
         annif.util.atomic_save(self._model, self.datadir, self.MODEL_FILE)
 
-    def _learn(self, corpus, params):
+    def _learn(
+        self,
+        corpus: DocumentCorpus,
+        params: dict[str, Any],
+    ) -> None:
         self.initialize()
         self._fit_model(
             corpus, int(params["learn-epochs"]), int(params["lmdb_map_size"])
         )
+
+    def get_model_metadata(self, model_filename: str) -> dict | None:
+        """Read metadata from Keras model files."""
+
+        try:
+            with zipfile.ZipFile(model_filename, "r") as zip:
+                with zip.open("metadata.json") as metadata_file:
+                    metadata_str = metadata_file.read().decode("utf-8")
+                    metadata = json.loads(metadata_str)
+                    return metadata
+        except Exception:
+            self.warning(f"Failed to read metadata from {model_filename}")
+            return None
