@@ -3,7 +3,10 @@ learns which concept suggestions from each backend are trustworthy using the
 PAV algorithm, a.k.a. isotonic regression, to turn raw scores returned by
 individual backends into probabilities."""
 
+from __future__ import annotations
+
 import os.path
+from typing import TYPE_CHECKING, Any
 
 import joblib
 import numpy as np
@@ -15,7 +18,11 @@ import annif.util
 from annif.exception import NotInitializedException, NotSupportedException
 from annif.suggestion import SubjectSuggestion, SuggestionBatch
 
-from . import backend, ensemble
+from . import ensemble
+
+if TYPE_CHECKING:
+    from annif.corpus.document import DocumentCorpus
+    from annif.project import AnnifProject
 
 
 class PAVBackend(ensemble.BaseEnsembleBackend):
@@ -30,12 +37,7 @@ class PAVBackend(ensemble.BaseEnsembleBackend):
 
     DEFAULT_PARAMETERS = {"min-docs": 10}
 
-    def default_params(self):
-        params = backend.AnnifBackend.DEFAULT_PARAMETERS.copy()
-        params.update(self.DEFAULT_PARAMETERS)
-        return params
-
-    def initialize(self, parallel=False):
+    def initialize(self, parallel: bool = False) -> None:
         super().initialize(parallel)
         if self._models is not None:
             return  # already initialized
@@ -53,23 +55,30 @@ class PAVBackend(ensemble.BaseEnsembleBackend):
                     backend_id=self.backend_id,
                 )
 
-    def _get_model(self, source_project_id):
+    def _get_model(self, source_project_id: str) -> dict[int, IsotonicRegression]:
         self.initialize()
         return self._models[source_project_id]
 
-    def _merge_source_batches(self, batch_by_source, sources, params):
+    def _merge_source_batches(
+        self,
+        batch_by_source: dict[str, SuggestionBatch],
+        sources: list[tuple[str, float]],
+        params: dict[str, Any],
+    ) -> SuggestionBatch:
         reg_batch_by_source = {}
         for project_id, batch in batch_by_source.items():
             reg_models = self._get_model(project_id)
             pav_batch = [
                 [
-                    SubjectSuggestion(
-                        subject_id=sugg.subject_id,
-                        score=reg_models[sugg.subject_id].predict([sugg.score])[0],
-                    )
-                    if sugg.subject_id in reg_models
-                    else SubjectSuggestion(
-                        subject_id=sugg.subject_id, score=sugg.score
+                    (
+                        SubjectSuggestion(
+                            subject_id=sugg.subject_id,
+                            score=reg_models[sugg.subject_id].predict([sugg.score])[0],
+                        )
+                        if sugg.subject_id in reg_models
+                        else SubjectSuggestion(
+                            subject_id=sugg.subject_id, score=sugg.score
+                        )
                     )  # default to raw score
                     for sugg in result
                 ]
@@ -82,7 +91,9 @@ class PAVBackend(ensemble.BaseEnsembleBackend):
         return super()._merge_source_batches(reg_batch_by_source, sources, params)
 
     @staticmethod
-    def _suggest_train_corpus(source_project, corpus):
+    def _suggest_train_corpus(
+        source_project: AnnifProject, corpus: DocumentCorpus
+    ) -> tuple[csc_matrix, csc_matrix]:
         # lists for constructing score matrix
         data, row, col = [], [], []
         # lists for constructing true label matrix
@@ -114,7 +125,9 @@ class PAVBackend(ensemble.BaseEnsembleBackend):
         )
         return csc_matrix(scores), csc_matrix(true)
 
-    def _create_pav_model(self, source_project_id, min_docs, corpus):
+    def _create_pav_model(
+        self, source_project_id: str, min_docs: int, corpus: DocumentCorpus
+    ) -> None:
         self.info(
             "creating PAV model for source {}, min_docs={}".format(
                 source_project_id, min_docs
@@ -138,7 +151,12 @@ class PAVBackend(ensemble.BaseEnsembleBackend):
             pav_regressions, self.datadir, model_filename, method=joblib.dump
         )
 
-    def _train(self, corpus, params, jobs=0):
+    def _train(
+        self,
+        corpus: DocumentCorpus,
+        params: dict[str, Any],
+        jobs: int = 0,
+    ) -> None:
         if corpus == "cached":
             raise NotSupportedException(
                 "Training pav project from cached data not supported."
