@@ -1,8 +1,8 @@
 """Definitions for command-line (Click) commands for invoking Annif
 operations and printing the results to console."""
 
-
 import collections
+import importlib
 import json
 import os.path
 import re
@@ -17,16 +17,25 @@ import annif.corpus
 import annif.parallel
 import annif.project
 import annif.registry
-from annif import cli_util
-from annif.exception import NotInitializedException, NotSupportedException
+from annif import cli_util, hfh_util
+from annif.exception import (
+    NotInitializedException,
+    NotSupportedException,
+    OperationFailedException,
+)
 from annif.project import Access
+from annif.simplemma_util import detect_language
 from annif.util import metric_code
 
 logger = annif.logger
 click_log.basic_config(logger)
 
-cli = FlaskGroup(create_app=annif.create_app, add_version_option=False)
+create_app = annif.create_flask_app
+cli = FlaskGroup(
+    create_app=create_app, add_default_commands=False, add_version_option=False
+)
 cli = click.version_option(message="%(version)s")(cli)
+cli.params = [opt for opt in cli.params if opt.name not in ("env_file", "app")]
 
 
 @cli.command("list-projects")
@@ -43,20 +52,35 @@ def run_list_projects():
     for details.
     """
 
-    template = "{0: <25}{1: <45}{2: <10}{3: <7}"
-    header = template.format("Project ID", "Project Name", "Language", "Trained")
+    column_headings = (
+        "Project ID",
+        "Project Name",
+        "Vocabulary ID",
+        "Language",
+        "Trained",
+        "Modification time",
+    )
+    table = [
+        (
+            proj.project_id,
+            proj.name,
+            proj.vocab.vocab_id if proj.vocab_spec else "-",
+            proj.language,
+            str(proj.is_trained),
+            cli_util.format_datetime(proj.modification_time),
+        )
+        for proj in annif.registry.get_projects(min_access=Access.private).values()
+    ]
+    template = cli_util.make_list_template(column_headings, *table)
+    header = template.format(*column_headings)
     click.echo(header)
     click.echo("-" * len(header))
-    for proj in annif.registry.get_projects(min_access=Access.private).values():
-        click.echo(
-            template.format(
-                proj.project_id, proj.name, proj.language, str(proj.is_trained)
-            )
-        )
+    for row in table:
+        click.echo(template.format(*row))
 
 
 @cli.command("show-project")
-@click.argument("project_id")
+@cli_util.project_id
 @cli_util.common_options
 def run_show_project(project_id):
     """
@@ -70,12 +94,13 @@ def run_show_project(project_id):
     click.echo(f"Vocabulary:        {proj.vocab.vocab_id}")
     click.echo(f"Vocab language:    {proj.vocab_lang}")
     click.echo(f"Access:            {proj.access.name}")
+    click.echo(f"Backend:           {proj.backend.name}")
     click.echo(f"Trained:           {proj.is_trained}")
-    click.echo(f"Modification time: {proj.modification_time}")
+    click.echo(f"Modification time: {cli_util.format_datetime(proj.modification_time)}")
 
 
 @cli.command("clear")
-@click.argument("project_id")
+@cli_util.project_id
 @cli_util.common_options
 def run_clear_project(project_id):
     """
@@ -93,10 +118,8 @@ def run_list_vocabs():
     List available vocabularies.
     """
 
-    template = "{0: <20}{1: <20}{2: >10}  {3: <6}"
-    header = template.format("Vocabulary ID", "Languages", "Size", "Loaded")
-    click.echo(header)
-    click.echo("-" * len(header))
+    column_headings = ("Vocabulary ID", "Languages", "Size", "Loaded")
+    table = []
     for vocab in annif.registry.get_vocabs(min_access=Access.private).values():
         try:
             languages = ",".join(sorted(vocab.languages))
@@ -106,11 +129,19 @@ def run_list_vocabs():
             languages = "-"
             size = "-"
             loaded = False
-        click.echo(template.format(vocab.vocab_id, languages, size, str(loaded)))
+        row = (vocab.vocab_id, languages, str(size), str(loaded))
+        table.append(row)
+
+    template = cli_util.make_list_template(column_headings, *table)
+    header = template.format(*column_headings)
+    click.echo(header)
+    click.echo("-" * len(header))
+    for row in table:
+        click.echo(template.format(*row))
 
 
 @cli.command("load-vocab")
-@click.argument("vocab_id")
+@click.argument("vocab_id", shell_complete=cli_util.complete_param)
 @click.argument("subjectfile", type=click.Path(exists=True, dir_okay=False))
 @click.option("--language", "-L", help="Language of subject file")
 @click.option(
@@ -148,7 +179,7 @@ def run_load_vocab(vocab_id, language, force, subjectfile):
 
 
 @cli.command("train")
-@click.argument("project_id")
+@cli_util.project_id
 @click.argument("paths", type=click.Path(exists=True), nargs=-1)
 @click.option(
     "--cached/--no-cached",
@@ -192,7 +223,7 @@ def run_train(project_id, paths, cached, docs_limit, jobs, backend_param):
 
 
 @cli.command("learn")
-@click.argument("project_id")
+@cli_util.project_id
 @click.argument("paths", type=click.Path(exists=True), nargs=-1)
 @cli_util.docs_limit_option
 @cli_util.backend_param_option
@@ -214,7 +245,7 @@ def run_learn(project_id, paths, docs_limit, backend_param):
 
 
 @cli.command("suggest")
-@click.argument("project_id")
+@cli_util.project_id
 @click.argument(
     "paths", type=click.Path(dir_okay=False, exists=True, allow_dash=True), nargs=-1
 )
@@ -258,7 +289,7 @@ def run_suggest(
 
 
 @cli.command("index")
-@click.argument("project_id")
+@cli_util.project_id
 @click.argument("directory", type=click.Path(exists=True, file_okay=False))
 @click.option(
     "--suffix", "-s", default=".annif", help="File name suffix for result files"
@@ -288,9 +319,7 @@ def run_index(
         raise click.BadParameter(f'language "{lang}" not supported by vocabulary')
     backend_params = cli_util.parse_backend_params(backend_param, project)
 
-    documents = annif.corpus.DocumentDirectory(
-        directory, None, None, require_subjects=False
-    )
+    documents = annif.corpus.DocumentDirectory(directory, require_subjects=False)
     results = project.suggest_corpus(documents, backend_params).filter(limit, threshold)
 
     for (docfilename, _), suggestions in zip(documents, results):
@@ -305,7 +334,7 @@ def run_index(
 
 
 @cli.command("eval")
-@click.argument("project_id")
+@cli_util.project_id
 @click.argument("paths", type=click.Path(exists=True), nargs=-1)
 @click.option("--limit", "-l", default=10, help="Maximum number of subjects")
 @click.option("--threshold", "-t", default=0.0, help="Minimum score threshold")
@@ -397,12 +426,16 @@ def run_eval(
         ):
             eval_batch.evaluate_many(hit_sets[project_id], subject_sets)
 
-    template = "{0:<30}\t{1}"
+    template = "{0:<30}\t{1:{fmt_spec}}"
     metrics = eval_batch.results(
         metrics=metric, results_file=results_file, language=project.vocab_lang
     )
     for metric, score in metrics.items():
-        click.echo(template.format(metric + ":", score))
+        if isinstance(score, int):
+            fmt_spec = "d"
+        elif isinstance(score, float):
+            fmt_spec = ".04f"
+        click.echo(template.format(metric + ":", score, fmt_spec=fmt_spec))
     if metrics_file:
         json.dump(
             {metric_code(mname): val for mname, val in metrics.items()},
@@ -411,12 +444,28 @@ def run_eval(
         )
 
 
+@cli.command("run")
+@click.option("--host", type=str, default="127.0.0.1")
+@click.option("--port", type=int, default=5000)
+@click.option("--log-level")
+@click_log.simple_verbosity_option(logger, default="ERROR")
+def run_app(**kwargs):
+    """
+    Run Annif in server mode for development.
+    \f
+    The server is for development purposes only.
+    """
+    kwargs = {k: v for k, v in kwargs.items() if v is not None}
+    cxapp = annif.create_cx_app()
+    cxapp.run(**kwargs)
+
+
 FILTER_BATCH_MAX_LIMIT = 15
 OPTIMIZE_METRICS = ["Precision (doc avg)", "Recall (doc avg)", "F1 score (doc avg)"]
 
 
 @cli.command("optimize")
-@click.argument("project_id")
+@cli_util.project_id
 @click.argument("paths", type=click.Path(exists=True), nargs=-1)
 @click.option(
     "--jobs", "-j", default=1, help="Number of parallel jobs (0 means all CPUs)"
@@ -514,7 +563,7 @@ def run_optimize(project_id, paths, jobs, docs_limit, backend_param):
 
 
 @cli.command("hyperopt")
-@click.argument("project_id")
+@cli_util.project_id
 @click.argument("paths", type=click.Path(exists=True), nargs=-1)
 @click.option("--trials", "-T", default=10, help="Number of trials")
 @click.option(
@@ -551,5 +600,196 @@ def run_hyperopt(project_id, paths, docs_limit, trials, jobs, metric, results_fi
     click.echo("---")
 
 
+@cli.command("upload")
+@click.argument("project_ids_pattern", shell_complete=cli_util.complete_param)
+@click.argument("repo_id")
+@click.option(
+    "--token",
+    help="""Authentication token, obtained from the Hugging Face Hub.
+    Will default to the stored token.""",
+)
+@click.option(
+    "--revision",
+    help="""An optional git revision to commit from. Defaults to the head of the "main"
+    branch.""",
+)
+@click.option(
+    "--commit-message",
+    help="""The summary / title / first line of the generated commit.""",
+)
+@click.option(
+    "--modelcard/--no-modelcard",
+    default=True,
+    help="Update or create a Model Card with upload.",
+)
+@cli_util.common_options
+def run_upload(
+    project_ids_pattern, repo_id, token, revision, commit_message, modelcard
+):
+    """
+    Upload selected projects and their vocabularies to a Hugging Face Hub repository.
+    \f
+    This command zips the project directories and vocabularies of the projects
+    that match the given `project_ids_pattern` to archive files, and uploads the
+    archives along with the project configurations to the specified Hugging Face
+    Hub repository. An authentication token and commit message can be given with
+    options. If the README.md does not exist in the repository it is
+    created with default contents and metadata of the uploaded projects, if it exists,
+    its metadata are updated as necessary.
+    """
+    from huggingface_hub import HfApi
+    from huggingface_hub.utils import HfHubHTTPError, HFValidationError
+
+    projects = hfh_util.get_matching_projects(project_ids_pattern)
+    click.echo(f"Uploading project(s): {', '.join([p.project_id for p in projects])}")
+
+    commit_message = (
+        commit_message
+        if commit_message is not None
+        else f"Upload project(s) {project_ids_pattern} with Annif"
+    )
+
+    fobjs, operations = [], []
+    try:
+        fobjs, operations = hfh_util.prepare_commits(projects, repo_id, token)
+        api = HfApi()
+        api.create_commit(
+            repo_id=repo_id,
+            operations=operations,
+            commit_message=commit_message,
+            revision=revision,
+            token=token,
+        )
+    except (HfHubHTTPError, HFValidationError) as err:
+        raise OperationFailedException(str(err))
+    else:
+        if modelcard:
+            hfh_util.upsert_modelcard(repo_id, projects, token, revision)
+    finally:
+        for fobj in fobjs:
+            fobj.close()
+
+
+@cli.command("download")
+@click.argument("project_ids_pattern")
+@click.argument("repo_id")
+@click.option(
+    "--token",
+    help="""Authentication token, obtained from the Hugging Face Hub.
+    Will default to the stored token.""",
+)
+@click.option(
+    "--revision",
+    help="""
+    An optional Git revision id which can be a branch name, a tag, or a commit
+    hash.
+    """,
+)
+@click.option(
+    "--force",
+    "-f",
+    default=False,
+    is_flag=True,
+    help="Replace an existing project/vocabulary/config with the downloaded one",
+)
+@click.option(
+    "--trust-repo",
+    default=False,
+    is_flag=True,
+    help="Allow download from the repository even when it has no entries in the cache",
+)
+@cli_util.common_options
+def run_download(project_ids_pattern, repo_id, token, revision, force, trust_repo):
+    """
+    Download selected projects and their vocabularies from a Hugging Face Hub
+    repository.
+    \f
+    This command downloads the project and vocabulary archives and the
+    configuration files of the projects that match the given
+    `project_ids_pattern` from the specified Hugging Face Hub repository and
+    unzips the archives to `data/` directory and places the configuration files
+    to `projects.d/` directory. An authentication token and revision can be given with
+    options. If the repository hasn’t been used for downloads previously
+    (i.e., it doesn’t appear in the Hugging Face Hub cache on local system), the
+    `--trust-repo` option needs to be used.
+    """
+
+    hfh_util.check_is_download_allowed(trust_repo, repo_id)
+
+    project_ids = hfh_util.get_matching_project_ids_from_hf_hub(
+        project_ids_pattern, repo_id, token, revision
+    )
+    click.echo(f"Downloading project(s): {', '.join(project_ids)}")
+
+    vocab_ids = set()
+    for project_id in project_ids:
+        project_zip_cache_path = hfh_util.download_from_hf_hub(
+            f"projects/{project_id}.zip", repo_id, token, revision
+        )
+        hfh_util.unzip_archive(project_zip_cache_path, force)
+        config_file_cache_path = hfh_util.download_from_hf_hub(
+            f"{project_id}.cfg", repo_id, token, revision
+        )
+        vocab_ids.add(hfh_util.get_vocab_id_from_config(config_file_cache_path))
+        hfh_util.copy_project_config(config_file_cache_path, force)
+
+    for vocab_id in vocab_ids:
+        vocab_zip_cache_path = hfh_util.download_from_hf_hub(
+            f"vocabs/{vocab_id}.zip", repo_id, token, revision
+        )
+        hfh_util.unzip_archive(vocab_zip_cache_path, force)
+
+
+@cli.command("completion")
+@click.option("--bash", "shell", flag_value="bash")
+@click.option("--zsh", "shell", flag_value="zsh")
+@click.option("--fish", "shell", flag_value="fish")
+def run_completion(shell):
+    """Generate the script for tab-key autocompletion for the given shell. To enable the
+    completion support in your current bash terminal session run\n
+        source <(annif completion --bash)
+    """
+
+    if shell is None:
+        raise click.UsageError("Shell not given, try --bash, --zsh or --fish")
+
+    script = os.popen(f"_ANNIF_COMPLETE={shell}_source annif").read()
+    click.echo(f"# Generated by Annif {importlib.metadata.version('annif')}")
+    click.echo(script)
+
+
+@cli.command("detect-language")
+@click.argument("languages")
+@click.argument(
+    "paths", type=click.Path(dir_okay=False, exists=True, allow_dash=True), nargs=-1
+)
+def run_detect_language(languages, paths):
+    """
+    Detect the language of a single text document from standard input or for one or more
+    document file(s) given its/their path(s).
+    """
+
+    langs = tuple(languages.split(","))
+
+    def detect_language_and_show(text, languages):
+        try:
+            proportions = detect_language(text, languages)
+        except ValueError as e:
+            raise click.UsageError(e)
+        for lang, score in proportions.items():
+            if lang == "unk":
+                lang = "?"
+            click.echo(f"{lang}\t{score:.04f}")
+
+    if paths and not (len(paths) == 1 and paths[0] == "-"):
+        doclist = cli_util.open_text_documents(paths, docs_limit=None)
+        for doc, path in zip(doclist.documents, paths):
+            click.echo(f"Detected languages for {path}")
+            detect_language_and_show(doc.text, langs)
+    else:
+        text = sys.stdin.read()
+        detect_language_and_show(text, langs)
+
+
 if __name__ == "__main__":
     cli()
