@@ -1,56 +1,71 @@
 """Clases for supporting document corpora"""
 
+from __future__ import annotations
+
 import glob
 import gzip
 import os.path
 import re
 from itertools import islice
+from typing import TYPE_CHECKING
 
 import annif.util
 
 from .subject import SubjectSet
 from .types import Document, DocumentCorpus
 
+if TYPE_CHECKING:
+    from collections.abc import Iterator
+
+    from annif.corpus.subject import SubjectIndex
+
 logger = annif.logger
 
 
 class DocumentDirectory(DocumentCorpus):
     """A directory of files as a full text document corpus"""
 
-    def __init__(self, path, subject_index, language, require_subjects=False):
+    def __init__(
+        self,
+        path: str,
+        subject_index: SubjectIndex | None = None,
+        language: str | None = None,
+        require_subjects: bool = False,
+    ) -> None:
         self.path = path
         self.subject_index = subject_index
         self.language = language
         self.require_subjects = require_subjects
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[tuple[str, str] | tuple[str, None]]:
         """Iterate through the directory, yielding tuples of (docfile,
-        subjectfile) containing file paths. If there is no key file and
-        require_subjects is False, the subjectfile will be returned as None."""
+        subjectfile) containing file paths. If require_subjects is False, the
+        subjectfile will be returned as None."""
 
         for filename in sorted(glob.glob(os.path.join(self.path, "*.txt"))):
-            tsvfilename = re.sub(r"\.txt$", ".tsv", filename)
-            if os.path.exists(tsvfilename):
-                yield (filename, tsvfilename)
-                continue
-            keyfilename = re.sub(r"\.txt$", ".key", filename)
-            if os.path.exists(keyfilename):
-                yield (filename, keyfilename)
-                continue
-            if not self.require_subjects:
+            if self.require_subjects:
+                tsvfilename = re.sub(r"\.txt$", ".tsv", filename)
+                if os.path.exists(tsvfilename):
+                    yield (filename, tsvfilename)
+                    continue
+                keyfilename = re.sub(r"\.txt$", ".key", filename)
+                if os.path.exists(keyfilename):
+                    yield (filename, keyfilename)
+                    continue
+            else:
                 yield (filename, None)
 
     @property
-    def documents(self):
-        for docfilename, keyfilename in self:
+    def documents(self) -> Iterator[Document]:
+        for docfilename, subjfilename in self:
             with open(docfilename, errors="replace", encoding="utf-8-sig") as docfile:
                 text = docfile.read()
-            if keyfilename is None:
+            if subjfilename is None:
                 yield Document(text=text, subject_set=None)
                 continue
-            with open(keyfilename, encoding="utf-8-sig") as keyfile:
+            with open(subjfilename, encoding="utf-8-sig") as subjfile:
                 subjects = SubjectSet.from_string(
-                    keyfile.read(), self.subject_index, self.language
+                    subjfile.read(), self.subject_index, self.language
                 )
             yield Document(text=text, subject_set=subjects)
 
@@ -58,12 +73,12 @@ class DocumentDirectory(DocumentCorpus):
 class DocumentFile(DocumentCorpus):
     """A TSV file as a corpus of documents with subjects"""
 
-    def __init__(self, path, subject_index):
+    def __init__(self, path: str, subject_index: SubjectIndex) -> None:
         self.path = path
         self.subject_index = subject_index
 
     @property
-    def documents(self):
+    def documents(self) -> Iterator[Document]:
         if self.path.endswith(".gz"):
             opener = gzip.open
         else:
@@ -72,7 +87,7 @@ class DocumentFile(DocumentCorpus):
             for line in tsvfile:
                 yield from self._parse_tsv_line(line)
 
-    def _parse_tsv_line(self, line):
+    def _parse_tsv_line(self, line: str) -> Iterator[Document]:
         if "\t" in line:
             text, uris = line.split("\t", maxsplit=1)
             subject_ids = {
