@@ -1,5 +1,7 @@
 """Mixin class for types that need a data directory"""
 
+from __future__ import annotations
+
 import os
 import os.path
 
@@ -7,11 +9,11 @@ import os.path
 class DatadirMixin:
     """Mixin class for types that need a data directory for storing files"""
 
-    def __init__(self, datadir, typename, identifier):
+    def __init__(self, datadir: str, typename: str, identifier: str) -> None:
         self._datadir_path = os.path.join(datadir, typename, identifier)
 
     @property
-    def datadir(self):
+    def datadir(self) -> str:
         if not os.path.exists(self._datadir_path):
             try:
                 os.makedirs(self._datadir_path)
